@@ -2,12 +2,17 @@ import array
 import asyncio
 import logging
 import socket
+import sys
 from collections import deque
 from copy import copy
 from typing import Any, Optional
 
+if sys.version_info[:2] < (3, 11):
+    from async_timeout import timeout as asyncio_timeout
+else:
+    from asyncio import timeout as asyncio_timeout
+
 from .. import introspection as intr
-from .._private.unmarshaller import Unmarshaller
 from ..auth import Authenticator, AuthExternal
 from ..constants import (
     BusType,
@@ -21,8 +26,11 @@ from ..errors import AuthError
 from ..message import Message
 from ..message_bus import BaseMessageBus
 from ..service import ServiceInterface
+from .message_reader import build_message_reader
 from .proxy_object import ProxyObject
 
+NO_REPLY_EXPECTED_VALUE = MessageFlag.NO_REPLY_EXPECTED.value
+
 
 def _future_set_exception(fut: asyncio.Future, exc: Exception) -> None:
     if fut is not None and not fut.done():
@@ -89,20 +97,20 @@ class _MessageWriter:
                 _future_set_exception(self.fut, e)
             self.bus._finalize(e)
 
-    def buffer_message(self, msg: Message, future=None):
+    def buffer_message(self, msg: Message, future=None) -> None:
         self.messages.append(
             (
-                msg._marshall(negotiate_unix_fd=self.negotiate_unix_fd),
+                msg._marshall(self.negotiate_unix_fd),
                 copy(msg.unix_fds),
                 future,
             )
         )
 
-    def _write_without_remove_writer(self):
+    def _write_without_remove_writer(self) -> None:
         """Call the write callback without removing the writer."""
         self.write_callback(remove_writer=False)
 
-    def schedule_write(self, msg: Message = None, future=None):
+    def schedule_write(self, msg: Message = None, future=None) -> None:
         queue_is_empty = not self.messages
         if msg is not None:
             self.buffer_message(msg, future)
@@ -152,17 +160,17 @@ class MessageBus(BaseMessageBus):
     :vartype connected: bool
     """
 
+    __slots__ = ("_loop", "_auth", "_writer", "_disconnect_future")
+
     def __init__(
         self,
         bus_address: str = None,
         bus_type: BusType = BusType.SESSION,
         auth: Authenticator = None,
-        negotiate_unix_fd=False,
-    ):
-        super().__init__(bus_address, bus_type, ProxyObject)
-        self._negotiate_unix_fd = negotiate_unix_fd
-        self._loop = asyncio.get_event_loop()
-        self._unmarshaller = self._create_unmarshaller()
+        negotiate_unix_fd: bool = False,
+    ) -> None:
+        super().__init__(bus_address, bus_type, ProxyObject, negotiate_unix_fd)
+        self._loop = asyncio.get_running_loop()
 
         self._writer = _MessageWriter(self)
 
@@ -190,7 +198,15 @@ class MessageBus(BaseMessageBus):
 
         future = self._loop.create_future()
 
-        self._loop.add_reader(self._fd, self._message_reader)
+        self._loop.add_reader(
+            self._fd,
+            build_message_reader(
+                self._sock,
+                self._process_message,
+                self._finalize,
+                self._negotiate_unix_fd,
+            ),
+        )
 
         def on_hello(reply, err):
             try:
@@ -213,7 +229,7 @@ class MessageBus(BaseMessageBus):
         )
 
         self._method_return_handlers[hello_msg.serial] = on_hello
-        self._stream.write(hello_msg._marshall())
+        self._stream.write(hello_msg._marshall(False))
         self._stream.flush()
 
         return await future
@@ -257,7 +273,8 @@ class MessageBus(BaseMessageBus):
 
         super().introspect(bus_name, path, reply_handler)
 
-        return await asyncio.wait_for(future, timeout=timeout)
+        async with asyncio_timeout(timeout):
+            return await future
 
     async def request_name(
         self, name: str, flags: NameFlag = NameFlag.NONE
@@ -334,7 +351,7 @@ class MessageBus(BaseMessageBus):
             - :class:`Exception` - If a connection error occurred.
         """
         if (
-            msg.flags & MessageFlag.NO_REPLY_EXPECTED
+            msg.flags.value & NO_REPLY_EXPECTED_VALUE
             or msg.message_type is not MessageType.METHOD_CALL
         ):
             await self.send(msg)
@@ -355,7 +372,7 @@ class MessageBus(BaseMessageBus):
 
         return future.result()
 
-    def send(self, msg: Message):
+    def send(self, msg: Message) -> asyncio.Future:
         """Asynchronously send a message on the message bus.
 
         .. note:: This method may change to a couroutine function in the 1.0
@@ -415,24 +432,15 @@ class MessageBus(BaseMessageBus):
 
         return handler
 
-    def _message_reader(self):
-        try:
-            while True:
-                if self._unmarshaller.unmarshall():
-                    self._on_message(self._unmarshaller.message)
-                    self._unmarshaller = self._create_unmarshaller()
-                else:
-                    break
-        except Exception as e:
-            self._finalize(e)
-
-    async def _auth_readline(self):
+    async def _auth_readline(self) -> str:
         buf = b""
         while buf[-2:] != b"\r\n":
-            buf += await self._loop.sock_recv(self._sock, 2)
+            # The auth protocol is line based, so we can read until we get a
+            # newline.
+            buf += await self._loop.sock_recv(self._sock, 1024)
         return buf[:-2].decode()
 
-    async def _authenticate(self):
+    async def _authenticate(self) -> None:
         await self._loop.sock_sendall(self._sock, b"\0")
 
         first_line = self._auth._authentication_start(
@@ -454,9 +462,12 @@ class MessageBus(BaseMessageBus):
                 )
                 self._stream.flush()
             if response == "BEGIN":
+                # The first octet received by the server after the \r\n of the BEGIN command
+                # from the client must be the first octet of the authenticated/encrypted stream
+                # of D-Bus messages.
                 break
 
-    def disconnect(self):
+    def disconnect(self) -> None:
         """Disconnect the message bus by closing the underlying connection asynchronously.
 
         All pending  and future calls will error with a connection error.
@@ -467,12 +478,6 @@ class MessageBus(BaseMessageBus):
         except Exception:
             logging.warning("could not close socket", exc_info=True)
 
-    def _create_unmarshaller(self) -> Unmarshaller:
-        sock = None
-        if self._negotiate_unix_fd:
-            sock = self._sock
-        return Unmarshaller(self._stream, sock)
-
     def _finalize(self, err: Optional[Exception] = None) -> None:
         try:
             self._loop.remove_reader(self._fd)
